# -*- coding: utf-8 -*
# Copyright (c) 2017 Civic Knowledge. This file is licensed under the terms of the
# Revised BSD License, included in this distribution as LICENSE

"""

"""


from metapack.exc import MetapackError

class NotebookError(MetapackError):
    pass
